#include <list>

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

#include <stdio.h>


#ifdef dDOUBLE
#  define dsDrawBox dsDrawBoxD
#  define dsDrawSphere dsDrawSphereD
#  define dsDrawCylinder dsDrawCylinderD
#  define dsDrawCapsule dsDrawCapsuleD
#endif

struct Object_T {
  dBodyID body;
  dGeomID geom;
  dReal color[4];

  Object_T() : body(0), geom(0) {}

  void setColor(dReal r, dReal g, dReal b, dReal a=1.0f) {
    color[0] = r; color[1] = g; color[2] = b; color[3] = a;
  }
};
static std::list<Object_T*> cursor_list;
static std::list<Object_T*> object_list;

static dWorldID world;
static dSpaceID space;
static dJointGroupID contactgroup;

static dGeomID ground;

static Object_T* cursor;
static dReal cursor_angle = 0.0f;


static void start();
static void command(int cmd);
static void simLoop( int pause );
static void nearCallback(void* data, dGeomID o1, dGeomID o2);

static void createTile(dReal x, dReal y, dReal angle);
static void createBullet(dReal x, dReal y, dReal angle, dReal force);
static void deleteTileAll();
static bool deleteTile();
static void drawObject(const Object_T* obj);


void start()
{
  static float xyz[3] = {0.0f, -5.0f, 3.0f };
  static float hpr[3] = {90.0f, -15.0f, 0.0f};

  dsSetViewpoint( xyz, hpr );
}

void command(int cmd)
{
  static float moveSize = 0.5;

  if (cmd == 'h' || cmd == 'l' || cmd == 'j' || cmd == 'k') {
    // manipulate cursor
    const dReal* pos = dGeomGetPosition(cursor->geom);
    dReal xyz[3] = {pos[0], pos[1], pos[2]};
    dReal mv1 = moveSize * cos(cursor_angle*M_PI / 180.0);
    dReal mv2 = moveSize * sin(cursor_angle*M_PI / 180.0);

    switch (cmd) {
    case 'h': xyz[0] -= mv1; xyz[1] += mv2; break;
    case 'l': xyz[0] += mv1; xyz[1] -= mv2; break;
    case 'k': xyz[0] += mv2; xyz[1] += mv1; break;
    case 'j': xyz[0] -= mv2; xyz[1] -= mv1; break;
    }
    dGeomSetPosition(cursor->geom, xyz[0], xyz[1], xyz[2]);
  }
  else if (cmd == '+' || cmd == '-') {
    if (cmd == '+') moveSize += 0.1;
    if (cmd == '-') moveSize -= 0.1;
    printf( "move size = %.1f\n", moveSize );
  }
  else if (cmd == 'n' || cmd == 'm') {
    switch (cmd) {
    case 'n': cursor_angle -= (moveSize*10.0); break;
    case 'm': cursor_angle += (moveSize*10.0); break;
    }
    dMatrix3 R;
    dRFromAxisAndAngle(R, 0.0f, 0.0f, -1.0f, cursor_angle*M_PI/180.0f);
    dGeomSetRotation(cursor->geom, R);
  }
  else if (cmd == 'a') {
    const dReal* pos = dGeomGetPosition(cursor->geom);
    createTile(pos[0], pos[1], cursor_angle);

    dReal xyz[3] = {pos[0], pos[1], pos[2]};
    xyz[0] += 0.6 * sin(cursor_angle*M_PI / 180.0);
    xyz[1] += 0.6 * cos(cursor_angle*M_PI / 180.0);
    dGeomSetPosition(cursor->geom, xyz[0], xyz[1], xyz[2]);
  }
  else if (cmd == 'd') {
    deleteTileAll();
  }
  else if (cmd == 'x') {
    if(deleteTile()) {
      const dReal* pos = dGeomGetPosition(cursor->geom);
      dReal xyz[3] = {pos[0], pos[1], pos[2]};
      xyz[0] -= 0.6 * sin(cursor_angle*M_PI / 180.0);
      xyz[1] -= 0.6 * cos(cursor_angle*M_PI / 180.0);
      dGeomSetPosition(cursor->geom, xyz[0], xyz[1], xyz[2]);
    }
  }
  else if (cmd == 'b') {
    const dReal* pos = dGeomGetPosition(cursor->geom);
    createBullet(pos[0], pos[1], cursor_angle, 1000.0f);
  }
}

void deleteTileAll()
{
  std::list<Object_T*>::iterator it = object_list.begin();
  while (it != object_list.end()) {
    dBodyDestroy((*it)->body);
    dGeomDestroy((*it)->geom);
    delete (*it);
    ++it;
  }
  object_list.clear();
}

bool deleteTile()
{
  if (!object_list.empty()) {
    Object_T* obj = object_list.back();
    dBodyDestroy(obj->body);
    dGeomDestroy(obj->geom);
    delete obj;
    object_list.pop_back();
    return true;
  }
  return false;
}

void createTile(dReal x, dReal y, dReal angle)
{
  dReal size[3] = { 0.5f, 0.2f, 1.0f };

  Object_T *tile = new Object_T();
  object_list.push_back(tile);

  tile->setColor(0.2f, 0.2f, 0.2f, 1.0f);

  tile->body = dBodyCreate(world);
  tile->geom = dCreateBox(space, size[0], size[1], size[2]);
  dGeomSetBody(tile->geom, tile->body);

  dGeomSetPosition(tile->geom, x, y, 1.0);

  dMatrix3 R;
  dRFromAxisAndAngle(R, 0.0f, 0.0f, -1.0f, angle*M_PI/180.0f);
  dGeomSetRotation(tile->geom, R);

  dMass mass;
  dMassSetBoxTotal(&mass, 1.0, size[0], size[1], size[2]);
  dBodySetMass(tile->body, &mass);
}


void createBullet(dReal x, dReal y, dReal angle, dReal force)
{
  dReal radius = 0.1f;

  Object_T *bullet = new Object_T();
  object_list.push_back(bullet);

  bullet->setColor(1.0f, 0.0f, 0.0f, 1.0f);

  bullet->body = dBodyCreate(world);
  bullet->geom = dCreateSphere(space, radius);
  dGeomSetBody(bullet->geom, bullet->body);

  dGeomSetPosition(bullet->geom, x, y, 1.0);

  dMass mass;
  dMassSetSphereTotal(&mass, 3.0, radius);
  dBodySetMass(bullet->body, &mass);

  dReal fx = force * sin(angle*M_PI/180.0);
  dReal fy = force * cos(angle*M_PI/180.0);
  
  dBodyAddForce(bullet->body, fx, fy, 1.0f);
}


void drawObject(const Object_T* obj)
{
    const dGeomID &g( obj->geom );
    if (!g) { return; }

    dReal color[4];
    color[0] = obj->color[0];
    color[1] = obj->color[1];
    color[2] = obj->color[2];
    color[3] = obj->color[3];
    dsSetColorAlpha(color[0], color[1], color[2], color[3]);

    int type = dGeomGetClass(g);
    switch (type){
    case dBoxClass:
      {
        dReal sizes[3];
        dGeomBoxGetLengths(g, sizes);
        dsDrawBox(dGeomGetPosition(g), dGeomGetRotation(g), sizes);
      }
    break;

    case dSphereClass:
      {
        dsDrawSphere(dGeomGetPosition(g), dGeomGetRotation(g), dGeomSphereGetRadius(g));
      }
    break;

    case dCylinderClass:
      {
        dReal r, l;
        dGeomCylinderGetParams(g, &r, &l);
        dsDrawCylinder( dGeomGetPosition(g), dGeomGetRotation(g), r, l);
      }
    break;

    case dCapsuleClass:
      {
        dReal r, l;
        dGeomCapsuleGetParams(g, &r, &l);
        dsDrawCapsule( dGeomGetPosition(g), dGeomGetRotation(g), r, l);
      }
    } // switch ((*it)->type)
}

void simLoop( int pause )
{
  if (!pause) {
    dSpaceCollide(space, 0, &nearCallback);
    //dWorldStep( world, 0.01f );
    dWorldQuickStep( world, 0.01f );

    dJointGroupEmpty(contactgroup);
  }

  {// draw cursor
    std::list<Object_T*>::iterator it = cursor_list.begin();
    while (it != cursor_list.end()) {
      drawObject( *it );
      ++it;
    }
  }

  {// draw Object
    std::list<Object_T*>::iterator it = object_list.begin();
    while (it != object_list.end()) {
      drawObject( *it );
      ++it;
    }
  }
}


void nearCallback(void* data, dGeomID o1, dGeomID o2)
{
  const int MAX_CONTACTS = 4;
  dBodyID b1 = dGeomGetBody(o1);
  dBodyID b2 = dGeomGetBody(o2);

  if (b1 && b2 && dAreConnectedExcluding(b1, b2, dJointTypeContact))
    return;

  bool isGround = ((o1 == ground) || (o2 == ground));

  dContact contact[MAX_CONTACTS];
  for (int i=0; i<MAX_CONTACTS; i++) {
    contact[i].surface.mode = dContactBounce | dContactSoftCFM;
    contact[i].surface.mu = isGround ? dInfinity : 0.1;
    contact[i].surface.bounce = 0.3;
    contact[i].surface.bounce_vel = 0.05;
    contact[i].surface.soft_cfm = 0.00001;
  }

  int numc = dCollide(o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof(dContact));
  if (numc > 0) {
    for (int i=0; i<numc; i++) {
      dJointID c = dJointCreateContact(world, contactgroup, contact+i);
      dJointAttach(c, b1, b2);
    }
  }
}

int main(int argc, char* argv[])
{
  dsFunctions fn;
  fn.version = DS_VERSION;
  fn.start = &start;
  fn.step = &simLoop;
  fn.command = &command;
  fn.stop = 0;
  fn.path_to_textures = "../ode-0.11.1/drawstuff/textures";

  // init
  dInitODE();

  // world & space
  world = dWorldCreate();
  space = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0);

  // plane
  ground = dCreatePlane(space, 0.0f, 0.0f, 1.0f, 0.0f);
  dWorldSetGravity(world, 0.0f, 0.0f, -9.8f);

//  // box
//  {
//    Object_T *box = new Object_T();
//    object_list.push_back(box);
//
//    box->setColor(1.0, 1.0, 0.0, 1.0);
//
//    box->body = dBodyCreate(world);
//    dBodySetPosition(box->body, 0.0, 0.0, 3.0);
//
//    box->geom = dCreateBox(space, 1.0f, 1.0f, 1.0f);
//    dGeomSetBody(box->geom, box->body);
//  }
//

  // cursor
  {
    cursor = new Object_T();
    cursor_list.push_back(cursor);

    cursor->setColor(0.0, 0.0, 1.0, 0.5);

    cursor->body = dBodyCreate(world);
    //dBodySetDynamic(cursor->body);
    dBodySetKinematic(cursor->body);
    dBodySetPosition(cursor->body, 0.0f, 0.0f, 0.5f);

    cursor->geom = dCreateSphere(0, 0.1f);
    dGeomSetBody(cursor->geom, cursor->body); // binding

    dMatrix3 R;
    dRFromAxisAndAngle(R, 0.0f, 0.0f, -1.0f, cursor_angle*M_PI/180.0f);
    dGeomSetRotation(cursor->geom, R);

    { // allow
      Object_T *cursor_allow = new Object_T();
      cursor_list.push_back(cursor_allow);

      cursor_allow->setColor(1.0f, 0.0f, 0.0f, 0.5f);

      cursor_allow->geom = dCreateCylinder(0, 0.3f, 0.01f );
      dGeomSetBody(cursor_allow->geom, cursor->body); // binding

      dMatrix3 R;
      dRFromAxisAndAngle(R, 1.0f, 0.0f, 0.0, 90.0f*M_PI/180.0f);
      dGeomSetOffsetRotation(cursor_allow->geom, R);
      dGeomSetOffsetPosition(cursor_allow->geom, 0.0f, 0.5f, 0.0f);
    }
  }

  printf("\n");
  printf(" -- DOMINO -- \n");
  printf(" command usage : \n");
  printf("   'a' : create Tile at cursor poisition\n");
  printf("   'b' : create Bullet at cursor position/angle\n");
  printf("   'x' : delete One object\n");
  printf("   'd' : delete All Object\n");
  printf("   'h' : move cursor - left\n");
  printf("   'j' : move cursor - front\n");
  printf("   'k' : move cursor - back\n");
  printf("   'l' : move cursor - right\n");
  printf("   'm' : cursor angle - plus\n");
  printf("   'n' : cursor angle - minus\n");
  printf("   '+' : divide move - plus\n");
  printf("   '-' : divide move - minus\n");
  printf("\n");
  printf(" ----------- \n");
  printf("\n");

  //
  // simulation loop
  //
  dsSimulationLoop(argc, argv, 800, 600, &fn);
  
  dJointGroupDestroy(contactgroup);
  dWorldDestroy(world);
  dSpaceDestroy(space);
  dCloseODE();

  return 0;
}
